"""Predictor base class."""
from __future__ import annotations

from abc import ABC, abstractmethod

class Predictor(ABC):
    """Base class for all predictors."""
    @classmethod
    @abstractmethod
    def load(cls, filename: str) -> Predictor:
        """
        Loads a model.

        Args:
            filename: string.

        Returns:
            Predictor subclass.
        """
        pass

    @abstractmethod
    def save(self, filename: str) -> None:
        """
        Saves model to file.

        Args:
            filename: string.

        Returns:
            None.
        """
        pass


"""AlgoSquare Package API."""
from __future__ import annotations

import os
import json

from .api import ApiObject, api_get, api_post
from .common import upload_file
from .predictor import Predictor

class Package(ApiObject):
    """Package class."""
    @classmethod
    def upload(cls, file: str) -> Package:
        """
        Creates a new package.
        
        Args:
            file: .whl file.

        Returns:
            Package.
        """
        if not file.endswith('.whl'):
            raise ValueError('file must be a .whl file')

        if not os.path.isfile(file):
            raise FileNotFoundError('file does not exist')

        files = [dict(namespace = "package", filename = os.path.split(file)[-1], key = upload_file(file))]    
        return cls(api_post('packages/api', json=dict(files = files)))

    @classmethod
    def get(cls, package_id: str) -> Package:
        """
        Gets specific package.
        
        Args:
            package_id: string.

        Returns:
            Package.
        """
        return cls(api_get(f'packages/{package_id}/api'))

    def refresh(self) -> None:
        """reloads task."""
        self.update(api_get(f'packages/{self.package_id}/api'))

    def get_predictors(self) -> list[Predictor]:
        """
        Gets predictors for package.
        
        Returns:
            List of Predictors.
        """
        if self.status != 'ok':
            raise RuntimeError('status must be ok')

        return [Predictor(x) for x in api_get(f'predictors/package/{self.package_id}/api')]


"""Testing of tabular algos."""
import uuid
import time
import pickle
import os
import numpy as np
import pandas as pd
import datetime

from typing import Union

from .errors import TestError
from .targets import create_classification_target, create_regression_target
from ..base.tabular import TabularClassificationAlgo, TabularRegressionAlgo, TabularClassifier, TabularRegressor, is_input_metatype, extract_tabular_statics
from ..base.types import Metatype, is_classification_metatype, is_delimiter_metatype, get_delimiter
from ..base.errors import ValidationError
from ..metrics.metrics import get_metric

def create_tabular_inputs(input_metatypes: set[Metatype], n_rows: int = 1000, n_cols: int = 2, seed: int = 0) -> tuple[pd.DataFrame, list[Metatype]]:
    X_metatypes = []
    data = []
    index = range(n_rows)

    for metatype in input_metatypes:
        X_metatypes += [metatype for _ in range(n_cols)]

        rng = np.random.default_rng(seed = seed)

        if metatype == Metatype.NUMERICAL:
            x = create_regression_target(index, n_cols, seed = seed)
        elif metatype == Metatype.CATEGORICAL:
            x = create_classification_target(index, n_cols, ['red', 'green', 'blue', 'yellow'], seed = seed)
        elif metatype == Metatype.BINARY:
            x = create_classification_target(index, n_cols, ['bad', 'good'], seed = seed)
        elif metatype == Metatype.DATETIME:
            timestamps = pd.DataFrame(datetime.datetime(2022, 2, 22, 0, 0).timestamp() + rng.normal(0, 1e6, (n_rows, n_cols)))
            x = timestamps.applymap(lambda x: datetime.datetime.fromtimestamp(x).isoformat())
        elif metatype == Metatype.TIMESTAMP:
            x = pd.DataFrame(datetime.datetime(2022, 2, 22, 0, 0).timestamp() + rng.normal(0, 1e6, (n_rows, n_cols)))
        elif is_delimiter_metatype(metatype, strict = True):
            delim = get_delimiter(metatype)
            categories = ['dog','cat','dolphin','lion']
            idx = pd.DataFrame(rng.integers(1,len(categories),size=(1000, n_cols)))
            x = idx.applymap(lambda x: delim.join(categories[:x]))
        else:
            raise TestError('invalid metatype')

        data.append(x)

    X = pd.concat(data, axis = 1)
    X.columns = [f'X{i}' for i in range(X.shape[1])]
    return (X.astype(str), X_metatypes)

def load_tabular_data(algo: Union[type[TabularClassificationAlgo], type[TabularRegressionAlgo]], nan_fraction: float = 0.1):
    X, X_metatypes = create_tabular_inputs(algo.input_metatypes())

    rng = np.random.default_rng(seed = 0)
    X.mask(rng.uniform(size = X.shape) < nan_fraction, inplace=True)

    is_multioutput = algo.is_multioutput()
    if not isinstance(is_multioutput, bool):
        raise TestError('is_multioutput should return bool')

    n_cols = 3 if is_multioutput else 1

    if issubclass(algo, TabularClassificationAlgo):
        target_metatype = algo.target_metatype()
        categories = ['bad', 'good'] if target_metatype == Metatype.BINARY else ['red', 'green', 'blue', 'yellow']
        y = create_classification_target(X.index, n_cols, categories)
        y_metatypes = [target_metatype for x in range(n_cols)]
    elif issubclass(algo, TabularRegressionAlgo):
        y = create_regression_target(X.index, n_cols)
        y_metatypes = [Metatype.NUMERICAL for x in range(n_cols)]
    else:
        raise TypeError('invalid algo class')

    return (X, y, X_metatypes, y_metatypes)

def test_tabular_algo(algo_class: Union[type[TabularClassificationAlgo], type[TabularRegressionAlgo]], output_dir: str, timeout: int, nan_fraction: float) -> None:
    """
    Tests tabular algo.
    
    Args:
        algo_class: subclass of TabularClassificationAlgo or TabularRegressionAlgo.
        output_dir: path for temp files.   
        timeout: training time in seconds.
        nan_fraction: fraction of missing observations in inputs.

    Raises:
        TypeError, TestError
    """
    if not (issubclass(algo_class, TabularClassificationAlgo) or issubclass(algo_class, TabularRegressionAlgo)):
        raise TypeError('invalid algo class')

    extract_tabular_statics(algo_class)

    if issubclass(algo_class, TabularClassificationAlgo):
        metric = get_metric('accuracy')
    else:
        metric = get_metric('mean_squared_error')

    X, y, X_metatypes, y_metatypes = load_tabular_data(algo_class, nan_fraction)

    model = algo_class()

    t_start = time.time()
    model.fit(timeout, metric, X, y, X_metatypes, y_metatypes, X_unlabelled = X)
    fit_seconds = time.time() - t_start

    if fit_seconds > timeout:
        raise TestError(f'timeout exceeded by {fit_seconds - timeout} seconds')

    print(f'Elapsed time: {fit_seconds}, timeout: {timeout}')

    test_tabular_model(model, output_dir, X, y)

def test_tabular_model(model: Union[TabularClassifier, TabularRegressor], output_dir: str, X: pd.DataFrame, y: pd.DataFrame) -> None:
    """
    Tests tabular model.
    
    Args:
        model: TabularClassifier or TabularRegressor.
        output_dir: path for temp files.   
        X: DataFrame.
        y: DataFrame.

    Raises:
        TypeError, TestError
    """
    if not (isinstance(model, TabularClassifier) or isinstance(model, TabularRegressor)):
        raise TypeError('invalid model')

    if not isinstance(X, pd.DataFrame):
        raise TypeError('X must be a DataFrame')

    if not isinstance(y, pd.DataFrame):
        raise TypeError('y must be a DataFrame')

    pred = model.predict(X)
    if not isinstance(pred, pd.DataFrame):
        raise TestError('prediction must be DataFrame')

    if list(pred.columns) != list(y.columns):
        raise TestError('prediction columns must equal y columns')

    if list(pred.index) != list(X.index):
        raise TestError('prediction index must equal X index')

    if isinstance(model, TabularClassifier):
        prediction_methods = [x.name.lower() for x in model.prediction_methods()]
        for prediction_method in prediction_methods:
            if prediction_method != 'predict':
                pred_prediction_method = getattr(model, prediction_method)(X)
                if y.shape[1] == 1:
                    pred_prediction_method = [pred_prediction_method]

                for ppm, col in zip(pred_prediction_method, y):
                    _check_prediction(prediction_method, ppm, X, y[col], pred[col])

    if output_dir is not None:
        filename = os.path.join(output_dir, uuid.uuid4().hex + '.mdl')
        model.save(filename)

        pred1 = _load_predictions(model.__class__, filename, X)
        pred2 = _load_predictions(model.__class__, filename, X)

        if not pred1.equals(pred2):
            raise TestError('loaded model predictions differ')

        os.remove(filename)

    print(f'TESTING DONE: {type(model).__name__}')

def _load_predictions(model_class, filename, X, prediction_method = 'predict'):
    model = model_class.load(filename)
    return getattr(model, prediction_method)(X)

def _check_prediction(prediction_method, ppm, X, y, pred):
    if not isinstance(ppm, pd.DataFrame):
        raise TestError(f'{prediction_method} must return a DataFrame for univariate and list thereof for multivariate problems')

    if list(ppm.index) != list(X.index):
        raise TestError('prediction index must equal X index')

    if prediction_method == 'predict_proba':
        if (ppm < 0).any(axis = None):
            raise TestError(f'{prediction_method} probabilities cannot be negative')

        if (ppm > 1).any(axis = None):
            raise TestError(f'{prediction_method} probabilities cannot be greater than 1')

        if ((ppm.sum(axis=1) - 1).abs() > 1e-6).any():
            raise TestError(f'{prediction_method} probabilities must sum to 1 for each row')

    if prediction_method in ('predict_proba', 'decision_function'):
        idx = ppm.to_numpy().argmax(axis = 1)
        if ppm.columns[idx].tolist() != pred.tolist():
            raise TestError(f'argmax of {prediction_method} output should equal predicted class')

    y_unique = set(y)

    if ppm.shape[1] != len(y_unique) or set(ppm.columns) != y_unique:
        raise TestError(f'{prediction_method} must have a column for each class')





"""Algo base class."""
from abc import ABC, abstractmethod
from typing import Optional

from .errors import ValidationError

class Algo(ABC):
    """Base class for all algos."""
    @staticmethod
    def is_private() -> bool:
        """
        Describes the prediction_methods that the classifier supports.
        
        Returns: 
            Bool.

        Note:
            A private algorithm cannot generate any earnings.
        """
        return True

    @staticmethod
    def hourly_rate() -> Optional[int]:
        """
        Describes the prediction_methods that the classifier supports.
        
        Returns: 
            None for private algorithms, int otherwise.
        """
        return None


def extract_algo_statics(algo_class: type[Algo]) -> dict:
    """Extracts static information from algo subclasses.

    Args:
        algo_class: subclass of Algo.

    Returns: 
        dict.

    Raises:
        ValidationError
    """
    if not issubclass(algo_class, Algo):
        raise ValidationError('algo must be a subclass of Algo')

    statics = dict()
    is_private = algo_class.is_private()

    if not isinstance(is_private, bool):
        raise ValidationError('is_private must be a bool')

    statics['is_private'] = is_private

    if not is_private:
        hourly_rate = algo_class.hourly_rate()

        if not isinstance(hourly_rate, int):
            raise ValidationError('hourly rate must be an int')

        if hourly_rate <= 0: 
            raise ValidationError('hourly_rate must be positive')

        statics['hourly_rate'] = hourly_rate

    return statics


"""Metric base class."""
from abc import ABC, abstractmethod
from typing import Any, Optional, Union

from .types import Metatype, PredictionMethod, is_target_metatype

class Metric(ABC):
    """Base class for all Metrics."""
    @abstractmethod 
    def __init__(self, name: str, metatypes: Union[Metatype, set[Metatype]], prediction_method: PredictionMethod, greater_is_better: bool):
        """
        Metric base class.

        Args:
            name: string representation.
            metatypes: Metatype or set thereof.
            prediction_method: predict, predict_proba, decision_function.
            greater_is_better: whether Metric is a score function (higher is better), or loss function (lower is better).
        """
        if not isinstance(name, str):
            raise TypeError('name must be a string')
        self._name = name

        if isinstance(metatypes, Metatype):
            metatypes = {metatypes}

        if not isinstance(metatypes, set):
            raise TypeError('metatypes must be a set')

        for metatype in metatypes:
            if not is_target_metatype(metatype, strict = True):
                raise ValueError('metatypes must be a set of Metatypes')

        if not isinstance(prediction_method, PredictionMethod):
            raise TypeError('prediction_method must be a PredictionMethod')

        if Metatype.NUMERICAL in metatypes and prediction_method != PredictionMethod.PREDICT:
            raise ValueError('prediction_method must be predict for numreical-metatype')

        self._metatypes = metatypes
        self._prediction_method = prediction_method.name.lower()
        self._sign = -1 if greater_is_better else 1

    @abstractmethod
    def score(self, y: Any, pred: Any) -> float:
        """
        Invokes underlying score function.

        Args:
            y: targets.
            pred: model predictions.

        Returns:
            float.
        """
        pass

    def __str__(self):
        """Name of Metric."""
        return self._name

    def to_dict(self) -> dict:
        metatypes = [metatype.name.lower() for metatype in self._metatypes]
        return dict(name = self._name, prediction_method = self._prediction_method, sign = self._sign, metatypes = metatypes)

    def loss(self, y: Any, pred: Any) -> float:
        """
        Loss based on underlying score function where lower is better.

        Args:
            y: targets.
            pred: model predictions.

        Returns:
            float.
        """
        return self.score_to_loss(self.score(y, pred))

    def score_to_loss(self, score: float) -> float:
        """
        Converts score to loss.

        Args:
            score: float.

        Returns:
            float.
        """
        return self._sign * score

    def prediction_score(self, model: Any, X: Any, y: Any) -> float:
        """
        Combined prediction and scoring.

        Args:
            model: estimator with prediction method.
            X: inputs.
            y: targets.

        Returns:
            float.
        """
        return self.score(y, self.prediction(model, X))

    def prediction_loss(self, model: Any, X: Any, y: Any) -> float:
        """
        Combined prediction and loss.

        Args:
            model: estimator with prediction method.
            X: inputs.
            y: targets.

        Returns:
            float.
        """
        return self.loss(y, self.prediction(model, X))

    def prediction(self, model: Any, *args, **kwargs) -> Any:
        """
        Returns prediction of model.

        Args:
            model: implements prediction method.
            *args: arguments to prediction method
            **kwargs: keyword arguments to prediction method

        Returns:
            model prediction.
        """
        return getattr(model, self._prediction_method)(*args, **kwargs)

    def supports_metatype(self, metatype: Metatype) -> bool:
        """
        Checks if metric supports a given metatype.

        Args:
            metatype: Metatype

        Returns:
            bool.
        """
        return metatype in self._metatypes



"""Tabular algos and auxiliary functions."""

from abc import ABC, abstractmethod
from typing import Optional, Union

import pandas as pd

from .algo import Algo
from .predictor import Predictor
from .types import Metatype, PredictionMethod, is_classification_metatype, is_prediction_method
from .errors import ValidationError
from .metric import Metric

INPUT_METATYPES = {Metatype.NUMERICAL, Metatype.CATEGORICAL, Metatype.BINARY,
            Metatype.DATETIME, Metatype.TIMESTAMP, Metatype.DELIM_PIPE, Metatype.DELIM_SEMICOLON}

class TabularClassifier(Predictor, ABC):
    """
    Base class for tabular classification model.

    Note:
        'predict_proba' and 'decision_function' are optional, but should be implemented if possible.
    """
    @staticmethod
    @abstractmethod
    def prediction_methods():
        """
        Describes the prediction_methods that the classifier supports.
        
        Returns: 
            Set of PredictionMethods.
        """
        pass

    @abstractmethod
    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Makes prediction for given data.
        
        Args:
            X: DataFrame with shape (n_samples, n_features).

        Returns:
            DataFrame with index of X and columns of y.
        """
        pass

    def predict_proba(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Outputs class probabilities for given data.
        
        Args:
            X: DataFrame with shape (n_samples, n_features).

        Returns:
            DataFrame with index of X and columns of classes in y, or list thereof for multivariate problems.

        Raises:
            NotImplementedError.
        """
        raise NotImplementedError

    def decision_function(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Outputs decision function for given data.
        
        Args:
            X: DataFrame with shape (n_samples, n_features).

        Returns:
            DataFrame with index of X and columns of classes in y, or list thereof for multivariate problems.
        
        Raises:
            NotImplementedError.
        """
        return self.predict_proba(X)

class TabularClassificationAlgo(Algo, TabularClassifier, ABC):
    """Base class for tabular classification learning algorithm."""
    @abstractmethod
    def fit(self, timeout: int, metric: Metric, X: pd.DataFrame, y: pd.DataFrame, X_metatypes: list[Metatype], y_metatypes: list[Metatype], X_unlabelled: Optional[pd.DataFrame] = None):
        """
        Fits model on data.

        Args:
            timeout: time in seconds to fit the model.
            metric: Metric.
            X: DataFrame with shape (n_samples, n_features) and dtype=str.
            y: DataFrame with shape (n_samples, n_outputs) and dtype=str.
            X_metatypes: list of n_features Metatypes.
            y_metatypes: list of n_outputs Metatypes.
            X_unlabelled: DataFrame with shape (n_samples_unlabelled, n_features).
        """
        pass

    @staticmethod
    @abstractmethod
    def input_metatypes() -> set[Metatype]:
        """
        Describes the X_metatypes that the algo supports.
        
        Returns: 
            Set of Metatypes.
        """
        pass

    @staticmethod
    @abstractmethod
    def target_metatype() -> Metatype:
        """
        Describes the y_metatype that the algo supports.
        
        Returns: 
            Metatype.
        """
        pass

    @staticmethod
    @abstractmethod
    def is_multioutput() -> bool:
        """
        Returns true if the algo supports mulitple outputs.
        
        Returns: 
            Boolean.
        """
        pass

class TabularRegressor(Predictor, ABC):
    """Base class for tabular regression model."""
    @abstractmethod
    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Makes prediction for given data.
        
        Args:
            X: DataFrame with shape (n_samples, n_features).

        Returns:
            DataFrame with index of X and columns of y.
        """
        pass

class TabularRegressionAlgo(Algo, TabularRegressor, ABC):
    """Base class for tabular regression learning algorithm."""
    @abstractmethod
    def fit(self, timeout: int, metric: Metric, X: pd.DataFrame, y: pd.DataFrame, X_metatypes: list[Metatype], y_metatypes: list[Metatype], X_unlabelled: Optional[pd.DataFrame] = None):
        """
        Fits model on data.

        Args:
            timeout: time in seconds to fit the model.
            metric: Metric.
            X: DataFrame with shape (n_samples, n_features) and dtype=str.
            y: DataFrame with shape (n_samples, n_outputs) and dtype=str.
            X_metatypes: list of n_features Metatypes.
            y_metatypes: list of n_outputs Metatypes.
            X_unlabelled: DataFrame with shape (n_samples_unlabelled, n_features).
        """
        pass

    @staticmethod
    @abstractmethod
    def input_metatypes() -> set[Metatype]:
        """
        Describes the X_metatypes that the algo supports.
        
        Returns: 
            Set of Metatypes.
        """
        pass

    @staticmethod
    @abstractmethod
    def is_multioutput() -> bool:
        """
        Returns true if the algo supports mulitple outputs.
        
        Returns: 
            Boolean.
        """
        pass

def is_input_metatype(metatype: Union[Metatype, int], strict: bool = False) -> bool:
    """Checks if metatype is valid input metatype.

    Args:
        metatype: Metatype or int.
        strict: checks if metatype is Metatype.

    Returns: 
        bool.
    """
    if strict and not isinstance(metatype, Metatype):
        return False

    try:
        metatype = Metatype(metatype)
        return metatype in INPUT_METATYPES
    except ValueError:
        return False

def extract_tabular_statics(tabular_class: Union[type[TabularClassifier], type[TabularRegressor]]) -> dict:
    """Extracts static information from tabular subclasses.

    Args:
        tabular_class: subclass of tabular classifier or regerssor.

    Returns: 
        dict.

    Raises:
        ValidationError
    """
    statics = dict()
    if issubclass(tabular_class, TabularClassificationAlgo) or issubclass(tabular_class, TabularRegressionAlgo):
        input_metatypes = tabular_class.input_metatypes()
        if not isinstance(input_metatypes, set) or not input_metatypes:
            raise ValidationError('input_metatypes must be a non-empty set')

        for metatype in input_metatypes:
            if not is_input_metatype(metatype, True):
                raise ValidationError('invalid input metatype')

        statics['input_metatypes'] = sorted([x.name.lower() for x in input_metatypes])

        is_multioutput = tabular_class.is_multioutput()
        if not isinstance(is_multioutput, bool):
            raise ValidationError('is_multioutput should return bool')

        statics['is_multioutput'] = is_multioutput

        if issubclass(tabular_class, TabularClassificationAlgo):
            target_metatype = tabular_class.target_metatype()
            if not is_classification_metatype(target_metatype, True):
                raise ValidationError('invalid classification metatype')

            statics['target_metatype'] = target_metatype.name.lower()


    if issubclass(tabular_class, TabularClassifier):
        prediction_methods = tabular_class.prediction_methods()
        if not isinstance(prediction_methods, set) or not prediction_methods:
            raise ValidationError('prediction_methods must be a non-empty set')

        for prediction_method in prediction_methods:
            if not is_prediction_method(prediction_method, True):
                raise ValidationError('invalid prediction method')

        if PredictionMethod.PREDICT not in prediction_methods:
            raise ValidationError('predict missing from prediction method')

        if PredictionMethod.PREDICT_PROBA in prediction_methods and PredictionMethod.DECISION_FUNCTION not in prediction_methods:
            raise ValidationError('if predict_proba is in prediction_methods, then decision_function should also be')

    else:
        prediction_methods = { PredictionMethod.PREDICT }

    statics['prediction_methods'] = sorted([prediction_method.name.lower() for prediction_method in prediction_methods])

    return statics

    


"""Test algo functionality."""
import inspect

from types import ModuleType
from typing import Any, Union

from .errors import TestError
from .tabular import test_tabular_algo
from ..base.algo import Algo, extract_algo_statics
from ..base.tabular import TabularClassificationAlgo, TabularRegressionAlgo

def test_algo(algo_class: type[Algo], output_dir: str, timeout: int = 300, nan_fraction: float = 0.1) -> None:
    """
    Function for testing algo functionality.

    Args:
        algo_class: subclass of Algo.
        output_dir: local directory for saving model.
        timeout: search time in seconds.
        nan_fraction: probability of null values in data.

    Returns:
        None.
    """
    extract_algo_statics(algo_class)

    if issubclass(algo_class, TabularClassificationAlgo) or issubclass(algo_class, TabularRegressionAlgo):
        test_tabular_algo(algo_class, output_dir, timeout, nan_fraction)
    else:
        raise TestError('invalid algo class')

def test_package(package: ModuleType, output_dir: str, **kwargs: Union[int, float]) -> None:
    """
    Function for testing algos in package.

    Args:
        package: module.
        output_dir: local directory for saving model.
        kwargs: keyword args passed to test_algo

    Returns:
        None.
    """
    for field in dir(package):
        attr = getattr(package, field)
        if inspect.isclass(attr) and issubclass(attr, Algo):
            test_algo(attr, output_dir, **kwargs)



"""Processing of tabular data."""
import datetime
import numpy as np
import pandas as pd
import hashlib

from typing import Optional

from ..base.types import Datatype, Metatype, DELIMITER_METATYPES, get_delimiter, is_target_metatype, is_classification_metatype
from ..base.errors import ValidationError

def process_table(data: pd.DataFrame, namespace: Optional[str] = None, metatypes: Optional[dict[Metatype]] = None, include: list[str] = None) -> list[dict]:
    """
    Create metadata for table.

    Args:
        data: DataFrame.
        namespace: string.
        metatypes: dictionary with name and Metatypes.
        include: only compute metadata for specified column names.

    Returns:
        list of dicts.

    Raises:
        ValidationError
    """
    if data.columns.unique().size != data.shape[1]:
        raise ValidationError('columns must be unique')

    output = []
    for name, series in data.items():
        if include is None or name in include:
            datatype, s = _cast_series(series.dropna())

            if datatype == Datatype.VOID:
                metatype = Metatype.VOID
            else:
                if metatypes is None or name not in metatypes:
                    metatype = _infer_metatype(s, name, datatype, namespace)
                else:
                    metatype = metatypes[name]

            metadata = dict(name = name, dtype = datatype, metatype = metatype, feature_id = hashlib.md5(str(name).encode()).hexdigest())

            if metatype != Metatype.VOID:
                _check_metatype(s, name, datatype, metatype, namespace)
                series_stats = _compute_series_stats(s, metatype)
                series_stats['missing'] = data.shape[0] - s.size
                metadata['stats'] = series_stats

            output.append(metadata)

    return output

def _cast_series(s):
    if s.isnull().any():
        raise ValueError('series has null values')

    if not s.size:
        return (Datatype.VOID, None)

    try:
        return (Datatype.INT, s.astype(int))
    except:
        pass

    try:
        return (Datatype.FLOAT, s.astype(float))
    except:
        pass

    return (Datatype.STR, s.astype(str))

def _infer_metatype(s, name, datatype, namespace = None):
    if datatype == Datatype.VOID:
        return Metatype.VOID

    if s.isnull().any():
        raise ValueError('series has null values')

    lowercase = name.lower()
    if lowercase == '' or 'unnamed:' in lowercase:
        return Metatype.VOID

    #void constants
    if not (s != s.iloc[0]).any():
        return Metatype.VOID

    num_unique = s.unique().size
    if datatype == Datatype.INT:
        if num_unique == 2:
            return Metatype.BINARY
        
        if lowercase.endswith('id') and (len(name) == 2 or name[-3] in '_-.:@' or name.endswith('Id')):
            return Metatype.CATEGORICAL

    if namespace == 'targets':
        if datatype in (Datatype.FLOAT, Datatype.INT):
            return Metatype.NUMERICAL
        return Metatype.BINARY if num_unique == 2 else Metatype.CATEGORICAL

    if datatype == Datatype.STR:
        if 'time' in lowercase or 'date' in lowercase:
            return Metatype.DATETIME

        for metatype in DELIMITER_METATYPES:
            delim = get_delimiter(metatype)
            for x in s:
                if delim in x:
                    return metatype

        return Metatype.BINARY if num_unique == 2 else Metatype.CATEGORICAL

    if 'timestamp' in name:
        return Metatype.TIMESTAMP

    return Metatype.NUMERICAL

def _check_metatype(s, name, datatype, metatype, namespace = None):
    if s.isnull().any():
        raise ValueError('series has null values')

    if metatype == Metatype.BINARY:
        if s.unique().size != 2:
            raise ValidationError(f'binary metatype {name} must contain exactly two values')

    if namespace == 'targets':
        if not is_target_metatype(metatype):
            raise ValidationError(f'{name} has invalid target-metatype')
        if datatype == Datatype.FLOAT and metatype != Metatype.NUMERICAL:
            raise ValidationError(f'{name} must be numerical metatype')
        if datatype == Datatype.STR and not is_classification_metatype(metatype):
            raise ValidationError(f'{name} must be binary or categorical metatype')
        if metatype == Metatype.CATEGORICAL:
            if s.unique().size == 2:
                raise ValidationError(f'categorical metatype {name} has two values and should be binary instead')

    try:
        if metatype == Metatype.NUMERICAL:
            s.astype(float)
        elif metatype == Metatype.DATETIME:
            [datetime.datetime.fromisoformat(x) for x in s]
        elif metatype == Metatype.TIMESTAMP:
            [datetime.datetime.fromtimestamp(x) for x in s]
    except:
        raise ValidationError(f'could not convert {name} to {metatype.name.lower()} metatype')

def _compute_series_stats(s, metatype):
    stats = dict()
    #casting ensures stats are json serializable
    if metatype == Metatype.NUMERICAL:
        stats['min'] = float(s.min())
        stats['max'] = float(s.max())
        stats['mean'] = float(s.mean())
        stats['std'] = float(s.std())
        stats['skew'] = float(s.skew())
        stats['kurtosis'] = float(s.kurtosis())
    elif metatype in (Metatype.BINARY, Metatype.CATEGORICAL):
        frequencies = s.value_counts(sort=False)

        if metatype == Metatype.BINARY:
            stats['categories'] = sorted(list(frequencies.index))
        else:
            stats['num_categories'] = int(frequencies.size)
        
        stats['min_count'] = int(frequencies.min())
        stats['max_count'] = int(frequencies.max())
        stats['gini'] = _gini(frequencies)

    return stats

def _gini(counts):
    nominator = 0
    for x in counts:
        for y in counts:
            nominator += np.abs(x-y)
    
    return nominator / (2 * len(counts) * sum(counts))


"""AlgoSquare Task API."""
from __future__ import annotations

from .api import ApiObject, api_get, api_post, api_delete
from .model import Model
from .algo import Algo

class Task(ApiObject):
    """Task class."""
    @classmethod
    def get(cls, task_id: str) -> Task:
        """
        Gets specific task.
        
        Args:
            task_id: string.

        Returns:
            Task.
        """
        return cls(api_get(f'tasks/{task_id}/api'))

    def refresh(self) -> None:
        """reloads task."""
        self.update(api_get(f'tasks/{self.task_id}/api'))

    def build_models(self, run_time: int = 60, max_algos: int = 10, server_cost: float = 0.2, max_algo_fee: float = 5, algo_ids: list[str] = []) -> None:
        """
        Builds model for task.
        
        Args:
            run_time: time in minutes.
            max_algos: maximum number of algos to use.
            server_cost: maximum server cost per hour in USD.
            max_algo_fee: maximum hourly rate of deployment.
            algo_ids: explicitly provide algos to use.

        Raises:
            RuntimeError, ValueError.
        """
        if self.status != 'inactive':
            raise RuntimeError('status must be inactive')

        run_time = int(run_time)
        if run_time < 5:
            raise ValueError('run_time must be minimum 5')

        payload = dict(run_time = run_time)

        if algo_ids:
            if len(algo_ids) > 20:
                raise ValueError('max 20 algos allowed')

            payload['algo_ids'] = algo_ids
        else:
            max_algos = int(max_algos)
            if max_algos < 1:
                raise ValueError('max_algos must be minimum 1')

            server_cost = float(server_cost)
            if server_cost < 0.2:
                raise ValueError('server_cost must be a minimum of 0.2')

            max_algo_fee = float(max_algo_fee)
            if max_algo_fee < 0:
                raise ValueError('max_algo_fee must not be negative')

            payload['max_algos'] = max_algos
            payload['server_cost'] = server_cost
            payload['max_algo_fee'] = max_algo_fee

        self.update(api_post(f'tasks/{self.task_id}/automl/api', json=payload))

    def stop_build(self) -> None:
        """
        Stops model building.
        
        Raises:
            RuntimeError.
        """
        if self.status != 'building_models':
            raise RuntimeError('status must be building_models')

        self.update(api_delete(f'tasks/{self.task_id}/automl/api'))

    def get_models(self) -> list[Model]:
        """
        Gets models for task.
        
        Returns:
            List of Models.
        """
        return [Model.load(x) for x in api_get(f'models/group/{self.task_id}/api')]

    def get_own_algos(self) -> list[Algo]:
        """
        Gets own algos for task.
        
        Returns:
            List of Algos.
        """
        return [Algo(x) for x in api_get(f'tasks/{self.task_id}/algos/api')]


